#!/usr/bin/perl 
use strict;
use warnings;

# COS 264 - Interactive Application Development 
# Spamfilter in Perl (Enhanced Version)
# Improved error handling, edge-case handling, consistent hash access, and better sorting.  
# Written by Chad Jordan (Improved and Extended from original 2009 version)
# 8-16-2015

use List::Util qw(sum);

my %good_msg = ();  
my %bad_msg = ();
my %odds = ();
my $num_bad = 0;
my $num_good = 0;
my %both_msg = ();

my $threshold = 0.9; 
my $min_word_occurrence = 5; 
my $default_prob = 0.4;
my $verbose = 1; 

#Common stop word search
my %stop_words = map { $_ => 1 } qw(
    the be to of and a in that have I it for not on with as you do at this but his by from
);

if (@ARGV < 3) {
    die "Usage: perl spamfilter.pl <spamfile> <goodfile> <unknownfile>\n";
}

#Convert lowercase and remove unwanted characters
sub normalize_word {
    my ($word) = @_;
    $word = lc($word); 
    $word =~ s/[^a-z0-9'\$-]//g; 
    return $word;
}

sub process_file {
    my ($filename, $hash_ref, $is_bad) = @_;
    open(my $fh, "<", $filename) or die "Cannot open file: $filename\n";
    my @lines = <$fh>;
    close $fh;
    
    foreach my $line (@lines) {
        if ($line =~ /^From:/) {
            $is_bad ? $num_bad++ : $num_good++;
        }
        my @words = split(/[^\w\-'\$]/, $line);
        foreach my $word (@words) {
            $word = normalize_word($word);
            next if $word eq '' || exists $stop_words{$word}; 
            $hash_ref->{$word}++;
        }
    }
}

process_file($ARGV[0], \%bad_msg, 1);
process_file($ARGV[1], \%good_msg, 0);

#Merging hashes for combined word list
%both_msg = (%good_msg, %bad_msg);

#Calculate probability
foreach my $word (keys %both_msg) {
    my $g = ($good_msg{$word} // 0) * 2;
    my $b = ($bad_msg{$word} // 0);
    
    if ($g + $b >= $min_word_occurrence) {
        my $freq_g = $num_good > 0 ? $g / $num_good : 0;
        my $freq_b = $num_bad > 0 ? $b / $num_bad : 0;
        
        $odds{$word} = $freq_b / ($freq_g + $freq_b) if ($freq_g + $freq_b) > 0;
        
        #Adjust for extreme odds
        $odds{$word} = 0.99 if $odds{$word} > 0.99;
        $odds{$word} = 0.01 if $odds{$word} < 0.01;
    }
}

my $unknown_file;
my %unknown_msg = ();
open($unknown_file, "<", $ARGV[2]) or die "Cannot open unknown file: $!";
my @unknown_lines = <$unknown_file>;
close $unknown_file;

#Regex sort for unknown probabilities
foreach my $line (@unknown_lines) {
    my @words = split(/[^\w\-'\$]/, $line);
    foreach my $word (@words) {
        $word = normalize_word($word);
        next if $word eq '' || exists $stop_words{$word};  
        $unknown_msg{$word} = exists $odds{$word} ? $odds{$word} : $default_prob;
    }
}

#Sort by relevance and probability
my @sort_words = sort {
    abs($unknown_msg{$a} - 0.5) <=> abs($unknown_msg{$b} - 0.5)
} keys %unknown_msg;
my @top_words = @sort_words[-15..-1];

if ($verbose) {
    print "Top Contributing Words:\n";
    foreach my $word (@top_words) {
        printf "%s: %.3f\n", $word, $unknown_msg{$word};
    }
}

my $prob_times = 1;
my $one_minus_probs = 1;
foreach my $word (@top_words) {
    $prob_times *= $unknown_msg{$word};
    $one_minus_probs *= (1 - $unknown_msg{$word});
}

if ($prob_times - $one_minus_probs == 0) {
    print "Unable to calculate probability. Not enough data.\n";
    exit;
}

my $answer = $prob_times / ($prob_times + $one_minus_probs);
printf("The Probability is: %.3f\n", $answer);

if ($answer > $threshold) {
    print "SPAM!\n";
} else {
    print "Not Spam\n";
}
